// -*-c++-*-
/* $Id: sfscd.h,v 1.45 1999/12/16 21:18:47 dm Exp $ */

/*
 *
 * Copyright (C) 1998, 1999 David Mazieres (dm@uun.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

#include "arpc.h"
#include "afsnode.h"
#include "nfsmounter.h"
#include "sfscd_prot.h"
#include "sfsmisc.h"
#include "crypt.h"
#include "rabin.h"
#include "qhash.h"
#include "afsroot.h"

extern int v3flag;
extern bool terminating;
extern bool nomounting;

extern ref<afsroot> afs_root;	// Files everyone sees in root directory
extern ref<afsdir> afs_sfsroot;	// Files seen by root during mount syscalls
extern ref<afsdir> afs_naroot;	// Files seen by users with no agents

extern ref<afsdir> afs_wait;	// Places people get "stuck" in readlink...
extern ref<afsdir> afs_mnt;	// ...to avoid locking the root directory

extern int afsfd;

void afs_init (cbv);

inline nfsmnt_handle
nfs_fh2tobytes (const nfs_fh &fh)
{
  nfsmnt_handle res;
  res.set (fh.data);
  return res;
}

extern bool root_is_nfsmounter;
void mnt_init ();
void mnt_mount (int s, str hostname, str path,
		int flags, const nfsmnt_handle &fh,
		callback<void, int, u_int64_t>::ref cb);
void mnt_remount (str path, int flags, cbi cb);
void mnt_umount (str path, int flags, cbi cb);
void mnt_umountall (int flags, cbi cb);

struct usrinfo : public virtual refcount {
  const sfs_aid aid;

  ref<afsusrdir> root;
  ihash_entry<usrinfo> hlink;
  ptr<axprt_unix> x;
  ptr<asrv> sh;
  ptr<aclnt> ch;
  bool clonelock;

  usrinfo (sfs_aid);
  ~usrinfo ();
  void dispatch (svccb *sbp);
  void authreq (sfscd_authreq *ar, svccb *sbp);
  void clonelock_clear () { clonelock = false; }
  static ptr<usrinfo> lookup (const authunix_parms *);
};
extern qhash<sfs_aid, ref<usrinfo> > &usrtab;

struct cdaemon;
struct afslink;

class srvinfo {
  typedef callback<void, srvinfo *, int>::ref alloccb_t;

  timecb_t *tmo;
  vector<alloccb_t> waitq;
  sfs_connectres conres;
  sfscd_mountres mntres;
  bool waiting;
  bool destroyed;
  bool cdmounted;
  bool visible_flag;

  srvinfo (const str &pathname, const str &hostname, const sfs_hash &hostid);
  PRIVDEST ~srvinfo ();

  void timeout (bool start = true);
  void fail (int);
  void connected (int);
  void gotconres (int, clnt_stat);
  void gotmntres (clnt_stat);
  void gotnfsmntres (int, u_int64_t);
  void ready ();
  void unmountcb (cbi::ptr, int);
  static void printdev (strbuf *sb, srvinfo *si);

public:
  u_int64_t devno;
  int error;

  const str path;
  const str hostname;
  const sfs_hash hostid;
  cdaemon *cdp;

  ihash_entry<srvinfo> hlink;
  list_entry<srvinfo> cdlink;

  void unmount (int flags, cbi::ptr cb = NULL);
  void destroy (bool stale);
  static srvinfo *lookup (const str &path);
  static void alloc (const str &path, alloccb_t cb);
  static void destroy (const str &path, cdaemon *cdp, bool stale);
  static void show (const str &path, cdaemon *cdp, bool showit);
  static void revoke (const str &path);
  static int pathstate (const str &path);
  static str devlist ();
};

class revocation : public afslink {
protected:
  revocation (const sfs_pathrevoke &c);
  ~revocation ();
  void update (const sfs_pathrevoke &c);

public:
  sfs_pathrevoke cert;
  sfs_hash hostid;
  ihash_entry<revocation> hlink;

  void nop () {}
  static ptr<revocation> alloc (const sfs_pathrevoke &c);
  static ptr<revocation> lookup (const str &path);
};

class cdaemon {
  void eof ();
  void dispatch (svccb *sbp);

public:
  str name;
  vector<str> argv;
  ptr<axprt_unix> x;
  ptr<aclnt> c;
  ptr<asrv> s;
  list<srvinfo, &srvinfo::cdlink> servers;
  ihash_entry<cdaemon> hlink;

  cdaemon (const vector<str> &argv);
  ~cdaemon ();

  bool launch (bool synchronous);
};
extern ihash<vector<str>, cdaemon, &cdaemon::argv, &cdaemon::hlink> &daemontab;
extern vector<sfs_extension> sfs_extensions;
extern bhash<in_addr> badaddrs;

class release {
  PRIVDEST ~release ();

public:
  struct prot {
    const u_int32_t prog;
    const u_int32_t vers;
    cdaemon *const cdp;
    ihash_entry<prot> link;
    prot (u_int32_t p, u_int32_t v, cdaemon *c)
      : prog (p), vers (v), cdp (c) {}
  };

  const u_int32_t rel;
  str libdir;
  ihash2<const u_int32_t, const u_int32_t, prot,
    &prot::prog, &prot::vers, &prot::link> prots;
  itree_entry<release> link;

  release (u_int32_t rel);
  static release *lookup (u_int32_t rel);
  static cdaemon *cdlookup (u_int32_t rel, u_int32_t prog, u_int32_t vers);
};

void update_devdb (cbv cb = cbv_null);

void usrinfo_init ();
void flushpath (str path);
sfs_aid cdgetaid (const authunix_parms *aup);

inline bool
isnfsmounter (const authunix_parms *aup)
{
  return cdgetaid (aup) == sfsaid_sfs;
}
